//twi.h

#define TW_MAX_RETRY    5

//twi define
#define TWI_FREQUENCY_HZ	        100000
#define TWI_READ 			1
#define TWI_WRITE 			0
#define TWI_GENERAL_ADDRESS	        0
#define TWSR_MASK 			0xF8

//TWCR register constants
#define TWINT		0b10000000
#define TWEA		0b01000000
#define TWSTA		0b00100000
#define TWSTO		0b00010000
#define TWWC		0b00001000
#define TWEN		0b00000100
#define TWIE		0b00000001

//twi master tx status codes                     
#define TWI_MT_ADR_ACK				0x18  	// SLA+W tramsmitted, ACK received
#define TWI_MT_ADR_NACK				0x20  	// SLA+W tramsmitted, NACK received 
#define TWI_MT_DATA_ACK				0x28  	// Data byte tramsmitted, ACK received
#define TWI_MT_DATA_NACK			0x30  	// Data byte tramsmitted, NACK received 

//twi master rx status codes  
#define TWI_MR_ADR_ACK				0x40  	// SLA+R tramsmitted, ACK received
#define TWI_MR_ADR_NACK				0x48  	// SLA+R transmitted, NACK received
#define TWI_MR_DATA_ACK				0x50  	// Data byte received, ACK tramsmitted
#define TWI_MR_DATA_NACK			0x58  	// Data byte received, NACK tramsmitted

//twi slave tx status codes
#define TWI_ST_ADR_ACK				0xA8  	// Own SLA+R received, ACK returned
#define TWI_ST_ADR_ACK_M_ARB_LOST	        0xB0  	// Arb lost in SLA+R/W as Master, own SLA+R received, ACK returned
#define TWI_ST_DATA_ACK				0xB8  	// Data byte TWDR transmitted, ACK received
#define TWI_ST_DATA_NACK			0xC0  	// Data byte TWDR transmitted, NACK received
#define TWI_ST_DATA_ACK_LAST_BYTE	        0xC8  	// Last data byte in TWDR transmitted, ACK received

//twi slave rx status codes
#define TWI_SR_ADR_ACK				0x60  	// Own SLA+W received, ACK returned
#define TWI_SR_ADR_ACK_M_ARB_LOST 	        0x68  	// Lost arb in SLA+R/W as Master, own SLA+W received, ACK returned
#define TWI_SR_GEN_ACK				0x70  	// General call addr received, ACK returned
#define TWI_SR_GEN_ACK_M_ARB_LOST 	        0x78  	// Lost arb in SLA+R/W as Master, general call addr received, ACK returned
#define TWI_SR_ADR_DATA_ACK			0x80  	// Previously addressed with own SLA+W, data received, ACK returned
#define TWI_SR_ADR_DATA_NACK		        0x88  	// Previously addressed with own SLA+W, data received, NACK returned
#define TWI_SR_GEN_DATA_ACK			0x90  	// Previously addressed with general call, data received, ACK returned
#define TWI_SR_GEN_DATA_NACK		        0x98  	// Previously addressed with general call, data received, NACK returned
#define TWI_SR_STOP_RESTART			0xA0  	// STOP condition or repeated START condition received as slave

//other status codes
#define TWI_START				0x08	// START transmitted  
#define TWI_REP_START				0x10	// Repeated START transmitted
#define TWI_ARB_LOST				0x38	// Arbitration lost
#define TWI_NO_STATE				0xF8	// No relevant state information
#define TWI_BUS_ERROR				0x00	// Bus error with illegal START/STOP conditions

//function return status
#define TWI_SUCCESS			        0 
#define TWI_INVALID_STATUS_ERROR	        1
#define TWI_NO_ACK_ERROR		        2
#define TWI_INVALID_SLA_ERROR		        3
#define TWI_ARB_LOST_ERROR		        4       
#define TWI_NO_DATA_ERROR		        5

//initialize
void initTwiMaster(unsigned long cpu_freq, unsigned char pullup_resistor) {
	float temp;
	
	if (pullup_resistor) {
		DDRC.0 = 0;
		DDRC.1 = 0;
		PORTC.0 = 1;
		PORTC.1 = 1;
	}		
	
	temp = (float) cpu_freq / (float) TWI_FREQUENCY_HZ;
        TWBR = (temp - 16) / 2 ;//TWI Bit Rate Register		
				//I2C Freq = CPU Freq / (16 + 2*TWBR*4^TWPS)						
	TWSR = 0x00;		//prescaler 1
}

//send START, send SLA+W, send register address, send START, send SLA+R, grab data, STOP 
unsigned char twiReadRegister(unsigned char slaveWriteAddress, unsigned char slaveReadAddress, 
                                unsigned int regAddress, unsigned char length, unsigned char *data) {
	unsigned char sla, i, n = 0;
	sla = slaveWriteAddress | (((regAddress >> 8) & 0x07) << 1);

	retry:
	if (n++ >= TW_MAX_RETRY)
		return TWI_NO_ACK_ERROR;
	begin:
	TWCR = TWINT | TWSTA | TWEN;	//START
	while (!(TWCR & TWINT));	//wait for tx (TWINT)
	switch (TWSR & TWSR_MASK) {	//check status
		case TWI_REP_START:
		case TWI_START:	        //OK
			break;
		case TWI_ARB_LOST:
			goto begin;
		default:
			return TWI_INVALID_STATUS_ERROR;
	}
	
	TWDR = sla | TWI_WRITE;		//send SLA+W
	TWCR = TWINT | TWEN;		//clear interrupt, start tx
	while (!(TWCR & TWINT));	//wait for tx (TWINT)
	switch (TWSR & TWSR_MASK) {	//check status
		case TWI_MT_ADR_ACK:	//OK
			break;
		case TWI_MT_ADR_NACK:	//NACK -> retry
			goto retry;
   		case TWI_ARB_LOST:
			goto begin;
		default:	        //error -> STOP
			goto error;
	}

	TWDR = regAddress;		//send register address
	TWCR = TWINT | TWEN;		//clear interrupt, start tx
	while (!(TWCR & TWINT));	//wait for tx (TWINT)
	switch (TWSR & TWSR_MASK) {	//check status
		case TWI_MT_DATA_ACK:	//OK -> break		
			break;
		case TWI_MT_DATA_NACK:	//NACK -> quit	
			goto quit;
   		case TWI_ARB_LOST:	
			goto begin;
		default:		//error -> STOP		
			goto error;
	}
	
	//Stop/pause TWI interface
	TWCR = 0; 
	delay_us(2);

	TWCR = TWINT | TWSTA | TWEN;	//send START
	while (!(TWCR & TWINT));	//wait for tx
	switch (TWSR & TWSR_MASK) {     //check status
		case TWI_START:
		case TWI_REP_START:
			break;
		case TWI_ARB_LOST:
			goto begin;
		default:
			goto error;
	}
    
	sla = slaveReadAddress |(((regAddress >> 8) & 0x07) << 1); 
	
	TWDR = sla | TWI_READ;		//send SLA+R
	TWCR = TWINT | TWEN;		//clear interrupt, start tx
	while (!(TWCR & TWINT));	//wait for tx
	switch (TWSR & TWSR_MASK) {     //check status
		case TWI_MR_ADR_ACK:
			break;
		case TWI_MR_ADR_NACK:
			goto quit;
		case TWI_ARB_LOST:
			goto begin;
		default:
			goto error;
	}
    
	for (i = 0; i < length; i++) {  //get content of register
		if (i == (length-1)) TWCR = TWINT | TWEN;
		else TWCR = TWINT | TWEN | TWEA;
		while (!(TWCR & TWINT));
		switch (TWSR & TWSR_MASK) {
			case TWI_MR_DATA_NACK:
			case TWI_MR_DATA_ACK:
				*data++ = TWDR;
				break;
			default:
				goto error;
		}
	}
	
	TWCR = TWINT | TWSTO | TWEN;    //STOP
	while(TWCR & TWSTO);
	return TWI_SUCCESS;             //return success status
	
	error:
	quit:
	TWCR = TWINT | TWSTO | TWEN;    //STOP
	while(TWCR & TWSTO);
	return TWI_INVALID_STATUS_ERROR;//return error status
}

//send START, send SLA+W, send register address, write data, STOP
unsigned char twiWriteRegister(unsigned char slaveWriteAddress, unsigned int regAddress, 
                                        unsigned char length, unsigned char *data) {
	unsigned char sla, i, n = 0;
	sla = slaveWriteAddress |(((regAddress >> 8) & 0x07) << 1);

	retry:
	if (n++ >= TW_MAX_RETRY)
		return TWI_NO_ACK_ERROR;
	begin:
	TWCR = TWINT | TWSTA | TWEN;	//send START
	while (!(TWCR & TWINT));        //wait for tx (TWINT)
	switch (TWSR & TWSR_MASK) {     //check status
		case TWI_REP_START:
		case TWI_START:		//OK
			break;
		case TWI_ARB_LOST:
			goto begin;
		default:
		return TWI_INVALID_STATUS_ERROR;        
	}

	TWDR = sla | TWI_WRITE;	        //send SLA+W
	TWCR = TWINT | TWEN;	        //clear interrupt, start tx
	while (!(TWCR & TWINT));        //wait for tx (TWINT)
	switch (TWSR & TWSR_MASK) {     //check status
		case TWI_MT_ADR_ACK:	//OK
			break;
		case TWI_MT_ADR_NACK:	//NACK -> retry
			goto retry;
   		case TWI_ARB_LOST:
			goto begin;
		default:	        //error -> STOP
			goto error;
	}
	
	TWDR = regAddress;		//send register address
	TWCR = TWINT | TWEN;	        //clear interrupt, start tx
	while (!(TWCR & TWINT));	//wait for tx (TWINT)
	switch (TWSR & TWSR_MASK) {     //check status
		case TWI_MT_DATA_ACK:	//OK		
			break;
		case TWI_MT_DATA_NACK:	//NACK -> quit	
			goto quit;
   		case TWI_ARB_LOST:
			goto begin;
		default:		//error -> STOP		
			goto error;
	}
	
	for (i = 0; i < length; i++) {  //write data
		TWDR = *data++;
		TWCR = TWINT | TWEN;
		while (!(TWCR & TWINT));
		switch (TWSR & TWSR_MASK) {     //check status
			case TWI_MT_DATA_ACK:	//OK		
				break;
			case TWI_MT_DATA_NACK:	//write protected? -> quit
				goto quit;
    		case TWI_ARB_LOST:
				goto begin;
			default:	        //error -> STOP		
				goto error;
		}
	}

	TWCR = TWINT | TWSTO | TWEN;	//STOP
	while(TWCR & TWSTO);
	return TWI_SUCCESS;
	
	error:
	quit:
	TWCR = TWINT | TWSTO | TWEN;	//STOP
	while(TWCR & TWSTO);
	return TWI_INVALID_SLA_ERROR;
}